import sys
from os import environ

from django.conf import settings
from django.core import management
from django.test.runner import DiscoverRunner

if sys.argv[1] == "test":
    # patch DiscoverRunner setup database to be able to create
    # dynamo db on startup
    original_setup_database = DiscoverRunner.setup_databases

    def custom_setup_database(*args, **kwargs):
        # we create dynamod db then call the normal setup database
        management.call_command('delete_dynamodb_table')
        management.call_command('create_dynamo_table', table_name=settings.DYNAMODB_TABLE)
        parallel = args[0].parallel
        if parallel > 1:
            for index in range(args[0].parallel):
                management.call_command('create_dynamodb_table', table_name_suffix=str(index + 1))

        return original_setup_database(*args, **kwargs)

    DiscoverRunner.setup_databases = custom_setup_database

    # path DiscoverRunner teardown to be able to tear down dynamodb
    original_teardown_databases = DiscoverRunner.teardown_databases

    def custom_teardown_database(*args, **kwargs):
        # here we delete dynamodb table
        management.call_command("delete_dynamodb_table")
        return original_teardown_databases(*args, **kwargs)

    DiscoverRunner.teardown_databases = custom_teardown_database

    # now we need to be able to truncate data from dynamo db
    # each time a test is being  executed.
    original_setup_class = DiscoverRunner.test_suite._handleClassSetUp


    def custom_handle_class_setup(*args, **kwargs):
        from ussd.store.journey_store.DynamoDb import dynamodb_table
        # here we delete all records in dynamodb
        actual_db_name = "test_" + "django-ussd-airflow"
        db_name = settings.DATABASES["default"]['NAME']
        actual_dynamo_db_name = environ["DYNAMODB_TABLE"]

        db_suffix = db_name.replace(actual_db_name, "")

        if db_suffix:
            dynamo_db_name = "test_" + actual_dynamo_db_name + db_suffix
        else:
            dynamo_db_name = "test_" + actual_dynamo_db_name

        settings.DYNAMODB_TABLE = dynamo_db_name

        table = dynamodb_table(
            settings.DYNAMODB_TABLE,
            "http://dynamodb:8000"
        )
        scan = table.scan()
        for item in scan['Items']:
            table.delete_item(Key={'id': item.get('id')})

        return original_setup_class(*args, **kwargs)


    DiscoverRunner.test_suite._handleClassSetUp = custom_handle_class_setup


from ..journey_store import JourneyStore
from collections import OrderedDict

store = OrderedDict()


class DummyStore(JourneyStore):
    """
    This store should not be used in production only meant to test the interface
    """

    def _get(self, name, version, screen_name, **kwargs):
        if version == 'edit_mode':
            return store.get('edit_mode', {}).get(name)
        if store.get(name):
            if version is None:  # get the latest journey created
                journey = store[name][next(reversed(store[name]))]
            else:
                journey = store[name].get(version)
            if screen_name is not None:
                return journey.get(screen_name)
            return journey
        return None

    def _all(self, name):
        return store.get(name, {})

    def _save(self, name, journey, version):
        if version == self.edit_mode_version:
            store['edit_mode'] = {name: journey}
        else:
            if store.get(name):
                store[name][version] = journey
            else:
                store[name] = OrderedDict({version: journey})
        return journey

    def _delete(self, name, version=None):
        if store.get(name):
            if store[name].get(version):
                del store[name][version]
            else:
                del store[name]

    def flush(self):
        pass




from ..journey_store import JourneyStore
import boto3
from structlog import get_logger
from django.conf import settings
from botocore.config import Config
from copy import deepcopy
from boto3.dynamodb.conditions import Key

BOTO_CORE_CONFIG = getattr(
    settings, 'BOTO_CORE_CONFIG', None)
USE_LOCAL_DYNAMODB_SERVER = getattr(
    settings, 'USE_LOCAL_DYNAMODB_SERVER', False)

logger = get_logger(__name__)


# We'll find some better way to do this.
_DYNAMODB_CONN = None
_DYNAMODB_TABLE = {}


# defensive programming if config has been defined
# make sure it's the correct format.
if BOTO_CORE_CONFIG:
    assert isinstance(BOTO_CORE_CONFIG, Config)

dynamo_kwargs = dict(
    service_name='dynamodb',
    config=BOTO_CORE_CONFIG
)


def dynamodb_connection_factory(low_level=False, endpoint=None):
    """
    Since SessionStore is called for every single page view, we'd be
    establishing new connections so frequently that performance would be
    hugely impacted. We'll lazy-load this here on a per-worker basis. Since
    boto3.resource.('dynamodb')objects are state-less (aside from security
    tokens), we're not too concerned about thread safety issues.
    """
    boto_config = deepcopy(dynamo_kwargs)
    if endpoint is not None:
        boto_config["endpoint_url"] = endpoint

    if low_level:
        return boto3.client(**boto_config)

    global _DYNAMODB_CONN

    if not _DYNAMODB_CONN:
        logger.debug("Creating a DynamoDB connection.")
        _DYNAMODB_CONN = boto3.resource(**boto_config)
    return _DYNAMODB_CONN


def dynamodb_table(table: str, endpoint=None):
    global _DYNAMODB_TABLE

    if not _DYNAMODB_TABLE.get(table):
        _DYNAMODB_TABLE[table] = dynamodb_connection_factory(endpoint=endpoint).Table(table)
    return _DYNAMODB_TABLE[table]


class DynamoDb(JourneyStore):
    edit_mode_version = "-1"
    journeyName = "journeyName"
    version = "version"

    def __init__(self, table_name, endpoint=None):
        self.table_name = table_name
        self.table = dynamodb_table(table_name, endpoint=endpoint)

    def _get(self, name, version, screen_name, **kwargs):
        screen_kwarg = {}
        if screen_name is not None:
            screen_kwarg["ProjectionExpression"] = screen_name

        if version is None:
            response = self.table.query(
                KeyConditionExpression=Key(self.journeyName).eq(name) & Key(self.version).gt(self.edit_mode_version),
                **kwargs
            )
            item = response.get("Items")[-1]
        else:
            key = {
                self.journeyName: name,
                self.version: version
            }

            response = self.table.get_item(Key=key, **screen_kwarg)
            item = response.get('Item')

        if item:
            if item.get(self.journeyName):
                del item[self.journeyName]
            if item.get(self.version):
                del item[self.version]

            if screen_name:
                item = item.get(screen_name)
        return item or None

    def _all(self, name):
        results = {}
        for i in self._query(name):
            version = i[self.version]
            del i[self.journeyName]
            del i[self.version]
            results[version] = i
        return results

    def _query(self, name, **kwargs):
        response = self.table.query(
            KeyConditionExpression=Key(self.journeyName).eq(name),
            **kwargs
        )
        return response.get('Items')

    def _save(self, name, journey, version):
        item = {
                self.journeyName: name,
                self.version: version
            }
        item.update(journey)

        response = self.table.put_item(
            Item=item
        )

    def _delete(self, name, version=None):
        items = [
            {
                self.journeyName: name,
                self.version: version
            }
        ]
        if version is None:
            items = self._query(name,
                                **{"ProjectionExpression": "{0}, {1}".format(self.journeyName, self.version)})

        with self.table.batch_writer() as batch:
            for i in items:
                batch.delete_item(
                    Key={
                        self.journeyName: i[self.journeyName],
                        self.version: i[self.version]
                    }
                )

    def flush(self):
        all_records = self.table.scan()
        for i in all_records['Items']:
            self._delete(i[self.journeyName], i[self.version])



from django.core.management import BaseCommand
from botocore.exceptions import ClientError
from ussd.store.journey_store.DynamoDb import dynamodb_connection_factory


class Command(BaseCommand):
    help = "delete dynamodb tables"

    def add_arguments(self, parser):
        parser.add_argument(
            '--delete_all', '--delete_all',
            default=False,
            help='Boolean ',
            action='store_true',
            dest='delete_all'
        )

    def handle(self, *args, **options):
        connection = dynamodb_connection_factory(low_level=True, endpoint="http://dynamodb:8000")
        tables = connection.list_tables().get('TableNames')

        for table_name in tables:
            if table_name.startswith("test") or options.get('delete_all'):
                try:
                    connection.delete_table(
                        TableName=table_name
                    )
                except ClientError as e:
                    if e.response['Error']['Code'] != \
                            'ResourceNotFoundException':
                        raise e

                if not options.get('ignore_logs'):
                    self.stdout.write(
                        "{0} dynamble table  deleted".format(table_name))



from django.conf.urls import url
from ussd.views import MermaidText, ValidateJourney

app_name = 'ussd'
urlpatterns = [
    url(r'mermaid_text$', MermaidText.as_view(), name="mermaid_text"),
    url(r'validate_journey$', ValidateJourney.as_view())
]



from django.core.management import BaseCommand
import subprocess
import sys


class Command(BaseCommand):
    help = "creates a session table if does not exist"

    def add_arguments(self, parser):
        parser.add_argument(
            '--table_name', '--table_name',
            default='journeyTable',
            help='String',
            dest='table_name'
        )

        parser.add_argument(
            '--table_name_suffix', '--table_name_suffix',
            default='',
            help='String',
            dest='table_name_suffix'
        )

    def handle(self, *args, **options):
        api_request_table = options.get('table_name')

        if sys.argv[1] == 'test':
            api_request_table = "test_" + api_request_table

        if options.get('table_name_suffix'):
            api_request_table += "_" + options.get('table_name_suffix')

        cloud_formation_parameters = "" \
                                     "WriteCapacityUnits=1 " \
                                     "DyanamoTableName={table} " \
                                     "PreRequisiteStack=None " \
                                     "RequestIndexReadCapacityUnits=1 " \
                                     "StateIndexReadCapacityUnits=1 " \
                                     "ReadCapacityUnits=1".format(table=api_request_table)

        command = 'dynamodb_cloud_formation_cli ' \
                  '--endpoint-url=http://dynamodb:8000 ' \
                  '--region=eu-west-1 --parameters="{parameters}" ' \
                  '/usr/src/app/devops/templates/cloud-formation/' \
                  'dynamodb.templates.yml ' \
                  '| sh'.format(parameters=cloud_formation_parameters)

        subprocess.check_call(['bash', '-c', command])



from django.test import TestCase
from django.core.exceptions import ValidationError
from copy import deepcopy
from ussd.core import UssdView
from ..DummyStore import DummyStore
from ..DynamoDb import DynamoDb
from django.core import management
from django.conf import settings


class TestDriverStore:
    class BaseDriverStoreTestCase(TestCase):

        maxDiff = None

        def setUp(self):
            # setup driver instance
            self.driver = self.setup_driver()

        def tearDown(self):
            self.driver.flush()

        def test_creating_invalid_journey(self):
            sample_journey = {
                "initial_screen": {
                    "type": "initial_screen",
                    "next_screen": "end_screen",
                    "default_language": "en"
                },
            }

            self.assertRaises(ValidationError,
                              self.driver.save,
                              name="journey_a",
                              journey=sample_journey,
                              version="0.0.2")

        def test_creating_journey(self):
            sample_journey = {
                "initial_screen": {
                    "type": "initial_screen",
                    "next_screen": "end_screen",
                    "default_language": "en"
                },
                "end_screen": {
                    "type": "quit_screen",
                    "text": "end screen"
                }
            }

            # Creating journey_a
            sample_journey_one = deepcopy(sample_journey)

            sample_journey_two = deepcopy(sample_journey_one)
            sample_journey_two['end_screen']['text'] = "end screen of sample journey two"

            sample_journey_three = deepcopy(sample_journey)
            sample_journey_three['end_screen']['text'] = "end screen of sample journey three"

            self.driver.save(name="journey_a", journey=sample_journey_one, version="0.0.1")
            self.driver.save(name="journey_a", journey=sample_journey_two, version="0.0.2")
            self.driver.save(name="journey_a", journey=sample_journey_three, version="0.0.3")


            # Creating journey_b
            sample_journey_b_one = deepcopy(sample_journey)
            sample_journey_b_one['end_screen']['text'] = "end screen of sample journey_b one"

            sample_journey_b_two = deepcopy(sample_journey_b_one)
            sample_journey_b_two['end_screen']['text'] = "end screen of sample journey_b two"

            sample_journey_b_three = deepcopy(sample_journey_b_one)
            sample_journey_b_three['end_screen']['text'] = "end screen of sample journey_b three"

            self.driver.save(name="journey_b", journey=sample_journey_b_one, version="0.0.1")
            self.driver.save(name="journey_b", journey=sample_journey_b_two, version="0.0.2")
            self.driver.save(name="journey_b", journey=sample_journey_b_three, version="0.0.3")

            # test getting the journey a
            a = sample_journey_two
            b = self.driver.get('journey_a', '0.0.2')
            self.assertDictEqual(sample_journey_two, self.driver.get('journey_a', '0.0.2'))
            self.assertEqual(sample_journey_three, self.driver.get('journey_a', '0.0.3'))
            self.assertEqual(sample_journey_one, self.driver.get('journey_a', '0.0.1'))

            # test getting the journey b
            self.assertEqual(sample_journey_b_two, self.driver.get('journey_b', '0.0.2'))
            self.assertEqual(sample_journey_b_three, self.driver.get('journey_b', '0.0.3'))
            self.assertEqual(sample_journey_b_one, self.driver.get('journey_b', '0.0.1'))

            # test that you cannot save/update with the same version
            self.assertRaises(ValidationError, self.driver.save, name="journey_a",
                              journey=sample_journey_two, version="0.0.3")

            # test getting all the journeys
            self.assertDictEqual(
                {"0.0.1": sample_journey_one, "0.0.2": sample_journey_two, "0.0.3": sample_journey_three},
                self.driver.all('journey_a'))

            # test getting the latest journey
            self.assertEqual(sample_journey_three, self.driver.get('journey_a'))

            # test getting part of screen alone
            self.assertEqual(sample_journey_three['end_screen'],
                             self.driver.get('journey_a', screen_name='end_screen'))
            self.assertEqual(sample_journey_two['end_screen'],
                             self.driver.get('journey_a', version="0.0.2", screen_name='end_screen'))

            # sanity check
            assert sample_journey_two['end_screen'] != sample_journey['end_screen']

            # testing deleting
            self.driver.delete('journey_a', version="0.0.1")
            self.assertIsNone(self.driver.get('journey_a', version="0.0.1"))

            # check that it hasn't deleted anything
            self.assertEqual(len(self.driver.all('journey_a')), 2)

            # deleting all journeys including the versions
            self.driver.delete('journey_a')
            self.assertEqual(len(self.driver.all('journey_a')), 0)

        def test_saving_journeys_thats_still_in_edit_mode(self):
            sample_journey = {
                "initial_screen": {
                    "type": "initial_screen",
                    "next_screen": "end_screen",
                    "default_language": "en"
                },
                "end_screen": {
                    "type": "quit_screen",
                    "text": "end screen"
                }
            }

            self.driver.save(name="journey_a", journey=sample_journey, version="0.0.1")

            self.assertEqual(sample_journey, self.driver.get('journey_a', '0.0.1'))

            sample_journey_edit_mode = {
                "initial_screen": {
                    "type": "initial_screen",
                    "next_screen": "end_screen",
                    "default_language": "en"
                },
            }

            # confirm its invalid
            is_valid, errors = UssdView.validate_ussd_journey(sample_journey_edit_mode)
            self.assertFalse(is_valid)

            # but we can still save it under edit mode
            self.driver.save(name='journey_a', journey=sample_journey_edit_mode, edit_mode=True)

            # stills get the valid one that was saved.
            self.assertEqual(sample_journey, self.driver.get('journey_a'))

            # getting the one use edit mode
            self.assertEqual(sample_journey_edit_mode, self.driver.get('journey_a', edit_mode=True))

            sample_journey_edit_mode['invalid_screen'] = "invalid_screen"

            # and we can still save it again
            self.driver.save(name='journey_a', journey=sample_journey_edit_mode, edit_mode=True)
            self.assertEqual(sample_journey_edit_mode, self.driver.get('journey_a', edit_mode=True))


class TestDummyStore(TestDriverStore.BaseDriverStoreTestCase):

    @staticmethod
    def setup_driver() -> DummyStore:
        return DummyStore()


class TestDynamodb(TestDriverStore.BaseDriverStoreTestCase):

    @staticmethod
    def setup_driver() -> DynamoDb:
        return DynamoDb(settings.DYNAMODB_TABLE, "http://dynamodb:8000")



"""
Customer journey are stored in a document store.
Any engine that implements this interface can be integrated with journey store.
"""
from ussd.core import UssdView
from django.core.exceptions import ValidationError
import abc


class JourneyStore(object, metaclass=abc.ABCMeta):

    edit_mode_version = "edit_mode"

    @abc.abstractmethod
    def _get(self, name, version, screen_name, **kwargs):
        pass

    @abc.abstractmethod
    def _all(self, name):
        pass

    @abc.abstractmethod
    def _save(self, name, journey, version):
        pass

    @abc.abstractmethod
    def _delete(self, name, version=None):
        pass

    @abc.abstractmethod
    def flush(self):
        pass

    def get(self, name: str, version=None, screen_name=None, edit_mode=False, propagate_error=True):
        if edit_mode:
            version = self.edit_mode_version
        return self._get(name, version, screen_name)

    def all(self, name: str):
        return self._all(name)

    def save(self, name: str, journey: dict, version=None, edit_mode=False):

        # version and editor mode should not be false
        if not (version or edit_mode):
            raise TypeError("version is required if its not in editor mode")

        if edit_mode:
            version = self.edit_mode_version

        # check if this version already exists
        if self.get(name, version, propagate_error=False) is not None:
            if not edit_mode:
                raise ValidationError("journey already exists")

        # validate if its not in editing mode.
        if not edit_mode:
            is_valid, errors = UssdView.validate_ussd_journey(journey)
            if not is_valid:
                raise ValidationError("invalid journey")

        # now create journey
        return self._save(name, journey, version)

    def delete(self, name, version=None):
        return self._delete(name, version)


